from os import environ
from contextlib import asynccontextmanager
from functools import wraps
from typing import List
from httpx import AsyncClient
from .api.v0 import aio_user
from .api.v0 import aio_models as aio_model0
from .api.v1 import aio_models as aio_model1
from .api.v0 import aio_prompt_completion as aio_prompt_completion0
from .api.v1 import aio_prompt_completion as aio_prompt_completion1
from .api.v0 import aio_file_upload
from .api.v0 import aio_image_generation, ImageSize
from httpx import RemoteProtocolError
from pathlib import Path
from .utils.models_to_enum import Model
from .utils import is_listable_not_string
from .api.v0_rag import (
    ChunkingMethod,
    BreakpointThresholdType,
    SearchType,
    aio_create_rag,
    aio_rags,
    aio_rag,
    aio_rag_delete,
    aio_rag_prompt_completion,
)
from .api.v0_agent import (
    aio_create_agent,
    aio_agents,
    aio_agent,
    aio_agent_prompt_completion,
    aio_add_rag_to_agent,
    aio_agent_delete,
    aio_agent_update,
)
from .async_client_agent import AsyncStraicoAgent
from .async_client_rag import AsyncStraicoRAG


def aio_retry_on_disconnect(func):
    @wraps(func)
    async def retry_func(self, *args, **kwargs):
        for i in range(1):
            try:
                r = await func(self, *args, **kwargs)
                return r
            except RemoteProtocolError as e:
                print("REconnect")
                await self._reconnect()

    return retry_func


class AsyncStraicoClient:
    def __init__(
        self, API_KEY: str = None, STRAICO_BASE_URL: str = None, **settings: dict
    ):
        self._client_settings = settings

        if API_KEY is None:
            API_KEY = environ.get("STRAICO_API_KEY")
            if API_KEY is None:
                raise Exception("Straico API Key is not defined")
            API_KEY = str(API_KEY)

        self._api_key = API_KEY

        if STRAICO_BASE_URL is None:
            STRAICO_BASE_URL = environ.get(
                "STRAICO_BASE_URL", "https://api.straico.com"
            )

        self.BASE_URL = STRAICO_BASE_URL

        self._session = AsyncClient()
        self.__prepare_header()

    async def _reconnect(self):
        await self._session.close()
        self._session = AsyncClient()

    def __prepare_header(self):
        self._header = {"Authorization": f"Bearer {self._api_key}"}

    @property
    def API_KEY(self):
        return self._api_key

    @API_KEY.setter
    def API_KEY(self, API_KEY):
        self._api_key = API_KEY
        self.__prepare_header()

    @aio_retry_on_disconnect
    async def user(self):
        response = await aio_user(
            self._session, self.BASE_URL, self._header, **self._client_settings
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def models(self, v=1):
        if v == 0:
            response = await aio_model0(
                self._session, self.BASE_URL, self._header, **self._client_settings
            )
        elif v == 1:
            response = await aio_model1(
                self._session, self.BASE_URL, self._header, **self._client_settings
            )
        else:
            raise Exception(f"Unsupported model api version {v}")

        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def prompt_completion(
        self,
        model,
        message,
        *,
        files: [Path | str] = [],
        images: [Path | str] = [],
        youtube_urls: [str] = [],
        temperature: float = None,
        max_tokens: float = None,
        display_transcripts=False,
        raw_output=False,
    ):

        if type(model) == dict and "model" in model:
            model = model["model"]
        elif type(model) == Model:
            model = model.model
        v = None
        if len(files) > 0 or len(youtube_urls) > 0 or len(images) > 0:
            if is_listable_not_string(files) and len(files) > 4:
                raise Exception(
                    f"Too many attached files. API is limited to 4 File attachments"
                )
            if is_listable_not_string(youtube_urls) and len(youtube_urls) > 4:
                raise Exception(
                    f"Too many youtube_urls files. API is limited to 4 Youtube URL attachments"
                )
            if is_listable_not_string(images) and len(images) > 4:
                raise Exception(
                    f"Too many image files. API is limited to 4 Image URL attachments"
                )
            v = 1
        if v is None:
            if isinstance(model, tuple) or isinstance(model, list):
                v = 1
                new_model = []
                for m in model:
                    if isinstance(m, dict) and "model" in m:
                        new_model.append(m["model"])
                    elif isinstance(m, Model):
                        new_model.append(m.model)
                    else:
                        new_model.append(m)
                model = new_model
            else:
                v = 0

        if v == 0:
            response = await aio_prompt_completion0(
                self._session,
                self.BASE_URL,
                self._header,
                model,
                message,
                temperature=temperature,
                max_tokens=max_tokens,
                **self._client_settings,
            )
        elif v == 1:
            if isinstance(youtube_urls, str):
                youtube_urls = [youtube_urls]

            if isinstance(files, str):
                files = [files]

            if isinstance(images, str):
                images = [images]

            file_urls = []
            for file in files:
                if isinstance(file, str):
                    if not file.strip().lower().startswith("http"):
                        file_path = Path(file)
                        if file_path.exists() and file_path.is_file():
                            file_url = await self.upload_file(file)
                        else:
                            raise Exception(f"Unknown file {file}")
                    else:
                        file_url = file
                elif isinstance(file, Path) and file.exists() and file.is_file():
                    file_url = await self.upload_file(file)
                else:
                    raise Exception(f"Unknown file type {type(file)} for file {file}")
                file_urls.append(file_url)

            images_urls = []
            for image in images:
                if isinstance(image, str):
                    if not image.strip().lower().startswith("http"):
                        file_path = Path(image)
                        if file_path.exists() and file_path.is_file():
                            image_url = await self.upload_file(image)
                        else:
                            raise Exception(f"Unknown file {image}")
                    else:
                        image_url = image
                elif isinstance(image, Path) and image.exists() and image.is_file():
                    image_url = await self.upload_file(image)
                else:
                    raise Exception(
                        f"Unknown file type {type(image)} for image {image}"
                    )
                images_urls.append(image_url)

            response = await aio_prompt_completion1(
                self._session,
                self.BASE_URL,
                self._header,
                model,
                message,
                file_urls=file_urls,
                images=images_urls,
                youtube_urls=youtube_urls,
                display_transcripts=display_transcripts,
                temperature=temperature,
                max_tokens=max_tokens,
                **self._client_settings,
            )
        if response.status_code == 201 and response.json()["success"]:
            if raw_output:
                return response.json()
            else:
                return response.json()["data"]

    @aio_retry_on_disconnect
    async def upload_file(self, file_to_upload: Path | str) -> str:
        if type(file_to_upload) == str:
            file_to_upload = Path(file_to_upload)
        # pdf, docx, pptx, txt, xlsx, mp3, mp4, html, csv, json
        content_type_mapping = {
            "mp3": "audio/mpeg",
            "mp4": "video/mp4",
            "pdf": "application/pdf",
            "docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            "pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            "txt": "text/plain",
            "xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "html": "text/html",
            # "htm": "text/html",
            "csv": "text/csv",
            "json": "application/json",
            "py": "text/x-python",
            "php": "application/x-httpd-php",
            "js": "application/javascript",
            "css": "text/css",
            "cs": "text/x-csharp",
            "swift": "text/x-swift",
            "kt": "text/x-kotlin",
            "xml": "application/xml",
            "ts": "application/typescript",
            "png": "image/png",
            "jpg": "image/jpeg",
            "jpeg": "image/jpeg",
            "webp": "image/webp",
            "gif": "image/gif",
        }
        if not file_to_upload.exists():
            raise Exception(f"Cannot find file {file_to_upload}")

        if not file_to_upload.is_file() or file_to_upload.is_dir():
            raise Exception(f"Not a FILE {file_to_upload}")

        with file_to_upload.open("rb") as binary_reader:
            content = binary_reader.read(-1)

        file_extension = file_to_upload.name.split(".")[-1].lower()

        content_type = content_type_mapping.get(file_extension)
        if content_type is None:
            raise Exception("Unsupported File Type")
        response = await aio_file_upload(
            self._session,
            self.BASE_URL,
            self._header,
            filename=file_to_upload.name,
            content_type=content_type,
            binary_data=content,
            **self._client_settings,
        )
        if response.status_code == 201 and response.json()["success"]:
            return response.json()["data"]["url"]

    #################################
    # Image Generation API
    ##############################
    @aio_retry_on_disconnect
    async def image_generation(
        self,
        model,
        description: str,
        size: ImageSize | str,
        variations: int,
        seed: int = None,
    ):
        if type(model) == dict and "model" in model:
            model = model["model"]
        elif type(model) == Model:
            model = model.model

        response = await aio_image_generation(
            self._session,
            self.BASE_URL,
            self._header,
            model=model,
            description=description,
            size=size,
            variations=variations,
            seed=seed,
            **self._client_settings,
        )
        if response.status_code == 201 and response.json()["success"]:
            return response.json()["data"]

    async def image_generation_as_zipfile(
        self,
        model,
        description: str,
        size: ImageSize | str,
        variations: int,
        destination_zip_path: Path | str,
        seed: int = None,
    ) -> Path:
        if type(destination_zip_path) == str:
            destination_zip_path = Path(destination_zip_path)

        image_details = await self.image_generation(
            model, description, size, variations, seed=seed
        )

        zip_url = image_details["zip"]

        response = await self._session.get(zip_url, **self._client_settings)
        content = response.read()

        if destination_zip_path.is_dir():
            zip_name = zip_url.split("/")[-1]
            destination_zip_path = destination_zip_path / zip_name

        with destination_zip_path.open("wb") as file_writer:
            file_writer.write(content)

        return destination_zip_path

    async def aclose(self):
        await self._session.aclose()

    async def image_generation_as_images(
        self,
        model,
        description: str,
        size: ImageSize | str,
        variations: int,
        destination_directory_path: Path | str,
        seed: int = None,
    ) -> [Path]:
        if type(destination_directory_path) == str:
            destination_directory_path = Path(destination_directory_path)

        if not destination_directory_path.is_dir():
            raise Exception("Destination path is not a directory")

        image_details = await self.image_generation(
            model, description, size, variations, seed=seed
        )

        image_urls = image_details["images"]
        image_paths = []
        for image_url in image_urls:
            response = await self._session.get(image_url, **self._client_settings)
            content = response.read()

            image_name = image_url.split("/")[-1]
            destination_image_path = destination_directory_path / image_name

            with destination_image_path.open("wb") as file_writer:
                file_writer.write(content)

            image_paths.append(destination_image_path)

        return image_paths

    #################################
    # RAG API
    ##############################
    @aio_retry_on_disconnect
    async def create_rag(
        self,
        name: str,
        description: str,
        *file_to_uploads: [Path | str],
        chunking_method: [ChunkingMethod | str] = None,
        chunk_size: int = 1000,
        chunk_overlap: int = 50,
        breakpoint_threshold_type: [
            BreakpointThresholdType | str
        ] = BreakpointThresholdType.percentile,
        buffer_size: int = 500,
        separator: [List[str] | str] = None,
    ) -> str:

        if len(file_to_uploads) > 4:
            raise Exception(
                "Too many files, Only accepts up to 4 Files per RAG Instance"
            )
        if len(file_to_uploads) == 0:
            raise Exception("Requires atleast 1 File per RAG Instance")

        response = await aio_create_rag(
            self._session,
            self.BASE_URL,
            self._header,
            name=name,
            description=description,
            chunking_method=chunking_method,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            breakpoint_threshold_type=breakpoint_threshold_type,
            buffer_size=buffer_size,
            separator=separator,
            files=file_to_uploads,
            **self._client_settings,
        )
        if response.status_code == 201 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def rags(self) -> str:
        response = await aio_rags(
            self._session,
            self.BASE_URL,
            self._header,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def rag(self, rag_id: str) -> str:
        response = await aio_rag(
            self._session,
            self.BASE_URL,
            self._header,
            rag_id,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def rag_delete(self, rag_id: str) -> str:
        response = await aio_rag_delete(
            self._session,
            self.BASE_URL,
            self._header,
            rag_id,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()

    @aio_retry_on_disconnect
    async def rag_prompt_completion(
        self,
        rag_id: str,
        model: str,
        message: str,
        search_type: [SearchType | str] = None,
        k: int = None,
        fetch_k: int = None,
        lambda_mult: float = None,
        score_threshold: float = None,
    ) -> str:

        if type(model) == dict and "model" in model:
            model = model["model"]
        elif type(model) == Model:
            model = model.model

        response = await aio_rag_prompt_completion(
            self._session,
            self.BASE_URL,
            self._header,
            rag_id,
            model,
            message,
            search_type,
            k,
            fetch_k,
            lambda_mult,
            score_threshold,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["response"]

    #################################
    # RAG object factory methods
    ##############################
    async def rag_object(self, rag_id):
        data = await self.rag(rag_id=rag_id)
        agent_obj = AsyncStraicoRAG(self, data)
        return agent_obj

    async def rag_objects(self):
        _rags = await self.rags()
        return [AsyncStraicoRAG(self, rag) for rag in _rags]

    async def new_rag(
        self,
        name: str,
        description: str,
        *file_to_uploads: [Path | str],
        chunking_method: [ChunkingMethod | str] = None,
        chunk_size: int = 1000,
        chunk_overlap: int = 50,
        breakpoint_threshold_type: [
            BreakpointThresholdType | str
        ] = BreakpointThresholdType.percentile,
        buffer_size: int = 500,
        separator: [List[str] | str] = None,
    ) -> str:
        _rag = await self.create_rag(
            name,
            description,
            *file_to_uploads,
            chunking_method=chunking_method,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            breakpoint_threshold_type=breakpoint_threshold_type,
            buffer_size=buffer_size,
            separator=separator,
        )
        return AsyncStraicoRAG(self, _rag)

    #################################
    # Agent API
    ##############################
    @aio_retry_on_disconnect
    async def create_agent(
        self,
        name: str,
        description: str,
        model: str,
        system_prompt: str,
        tags: [str] = [],
        rag: [AsyncStraicoRAG | dict | str] = None,
    ) -> str:

        if type(model) == dict and "model" in model:
            model = model["model"]
        elif type(model) == Model:
            model = model.model

        response = await aio_create_agent(
            self._session,
            self.BASE_URL,
            self._header,
            name=name,
            description=description,
            model=model,
            system_prompt=system_prompt,
            tags=tags,
            **self._client_settings,
        )
        if response.status_code == 201 and response.json()["success"]:
            _agent = response.json()["data"]
            if rag is not None:
                rag_type = type(rag)
                if rag_type == dict and "_id" in rag:
                    rag = rag["_id"]
                if rag_type == AsyncStraicoRAG:
                    rag = rag.data["_id"]
                return await self.agent_add_rag(_agent["_id"], rag)
            return _agent

    @aio_retry_on_disconnect
    async def agents(self, *, with_tag: str = None) -> str:
        response = await aio_agents(
            self._session,
            self.BASE_URL,
            self._header,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            _agents = response.json()["data"]
            if with_tag is None:
                return _agents
            return [agent for agent in _agents if with_tag in agent["tag"]]

    @aio_retry_on_disconnect
    async def agent(self, agent_id: str) -> str:
        response = await aio_agent(
            self._session,
            self.BASE_URL,
            self._header,
            agent_id,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def agent_delete(self, agent_id: str) -> dict:
        response = await aio_agent_delete(
            self._session,
            self.BASE_URL,
            self._header,
            agent_id,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()

    @aio_retry_on_disconnect
    async def agent_add_rag(
        self, agent_id: str, rag_id: [AsyncStraicoRAG | dict | str]
    ) -> dict:
        rag_type = type(rag_id)
        if rag_type == dict and "_id" in rag_id:
            rag_id = rag_id["_id"]
        elif rag_type == AsyncStraicoRAG:
            rag_id = rag_id.data["_id"]

        response = await aio_add_rag_to_agent(
            self._session,
            self.BASE_URL,
            self._header,
            agent_id,
            rag_id,
            **self._client_settings,
        )

        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    @aio_retry_on_disconnect
    async def agent_prompt_completion(
        self,
        agent_id: str,
        message: str,
        search_type: [SearchType | str] = None,
        k: int = None,
        fetch_k: int = None,
        lambda_mult: float = None,
        score_threshold: float = None,
    ) -> dict:

        response = await aio_agent_prompt_completion(
            self._session,
            self.BASE_URL,
            self._header,
            agent_id,
            message,
            search_type,
            k,
            fetch_k,
            lambda_mult,
            score_threshold,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["response"]

    @aio_retry_on_disconnect
    async def agent_update(
        self,
        agent_id: str,
        rag: [AsyncStraicoRAG | dict | str] = None,
        name: str = None,
        description: str = None,
        model: str = None,
        system_prompt: str = None,
        tags: [str] = None,
    ) -> str:

        if rag is not None:
            rag_type = type(rag)
            if rag_type == dict and "_id" in rag:
                rag = rag["_id"]
            elif rag_type == AsyncStraicoRAG:
                rag = rag.data["_id"]

        if type(model) == dict and "model" in model:
            model = model["model"]

        elif type(model) == Model:
            model = model.model

        response = await aio_agent_update(
            self._session,
            self.BASE_URL,
            self._header,
            agent_id,
            rag_id=rag,
            name=name,
            description=description,
            model=model,
            system_prompt=system_prompt,
            tags=tags,
            **self._client_settings,
        )
        if response.status_code == 200 and response.json()["success"]:
            return response.json()["data"]

    #################################
    # Agent object factory methods
    ##############################
    async def agent_object(self, agent_id):
        data = await self.agent(agent_id=agent_id)
        agent_obj = AsyncStraicoAgent(self, data)
        return agent_obj

    async def agent_objects(self, *, with_tag: str = None):
        _agents = await self.agents(with_tag=with_tag)
        return [AsyncStraicoAgent(self, agent) for agent in _agents]

    async def new_agent(
        self,
        name: str,
        description: str,
        model: str,
        system_prompt: str,
        tags: [str] = [],
        rag: [AsyncStraicoRAG | dict | str] = None,
    ) -> str:
        _agent = await self.create_agent(
            name=name,
            description=description,
            model=model,
            system_prompt=system_prompt,
            tags=tags,
            rag=rag,
        )
        return AsyncStraicoAgent(self, _agent)


@asynccontextmanager
async def aio_straico_client(
    API_KEY: str = None, STRAICO_BASE_URL: str = None, **settings: dict
):
    try:
        client = AsyncStraicoClient(
            API_KEY=API_KEY, STRAICO_BASE_URL=STRAICO_BASE_URL, **settings
        )
        yield client
    finally:
        await client.aclose()



import pytest
from .utils import make_temp_file
from aio_straico import straico_client, aio_straico_client
from aio_straico.api.v0_rag import ChunkingMethod


@pytest.mark.asyncio
async def test_basic_rag_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(rag_name, rag_description, file)
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == "fixed_size"
            assert s["chunk_size"] == 1000
            assert s["chunk_overlap"] == 50


@pytest.mark.asyncio
async def test_customize_rag_fixed_size_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(
                rag_name,
                rag_description,
                file,
                chunk_size=1001,
                chunk_overlap=55,
                separator="a",
            )
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == "fixed_size"
            assert s["chunk_size"] == 1001
            assert s["chunk_overlap"] == 55
            assert s["separator"] == "a"


@pytest.mark.asyncio
async def test_customize_rag_fixed_size_explicit_str_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(
                rag_name,
                rag_description,
                file,
                chunking_method="fixed_size",
                chunk_size=1001,
                chunk_overlap=55,
                separator="a",
            )
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == "fixed_size"
            assert s["chunk_size"] == 1001
            assert s["chunk_overlap"] == 55
            assert s["separator"] == "a"


@pytest.mark.asyncio
async def test_customize_rag_fixed_size_explicit_enum_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(
                rag_name,
                rag_description,
                file,
                chunking_method=ChunkingMethod.fixed_size,
                chunk_size=1001,
                chunk_overlap=55,
                separator="a",
            )
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == ChunkingMethod.fixed_size.value
            assert s["chunk_size"] == 1001
            assert s["chunk_overlap"] == 55
            assert s["separator"] == "a"


@pytest.mark.asyncio
async def test_customize_rag_recursive_explicit_str_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(
                rag_name,
                rag_description,
                file,
                chunking_method="recursive",
                chunk_size=1001,
                chunk_overlap=55,
                separator=["a"],
            )
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == "recursive"
            assert s["chunk_size"] == 1001
            assert s["chunk_overlap"] == 55
            assert s["separators"] == ["a"]


@pytest.mark.asyncio
async def test_customize_rag_recursive_explicit_enum_aio():
    rag_name = "testrag"
    rag_description = "this is only a basic test"
    test_filename = "test.py"
    async with aio_straico_client() as client:
        with make_temp_file(test_filename, "this is a test") as file:
            s = await client.create_rag(
                rag_name,
                rag_description,
                file,
                chunking_method=ChunkingMethod.recursive,
                chunk_size=1001,
                chunk_overlap=55,
                separator=["a"],
            )
            assert s["name"] == rag_name
            assert s["description"] == rag_description
            assert s["original_filename"] == test_filename
            assert s["chunking_method"] == ChunkingMethod.recursive.value
            assert s["chunk_size"] == 1001
            assert s["chunk_overlap"] == 55
            assert s["separators"] == ["a"]



class ModelCategory:
    pass


class ModelProvider:
    pass


class Model:
    def __init__(self, **properties):
        self.model = properties["model"]
        self.properties = properties

    def __getitem__(self, key):
        return self.properties.get(key)

    def __str__(self):
        return self.model

    def __repr__(self):
        return str(self.properties)


def __python_name(name):
    f = []
    for x in name:
        if x.isalnum():
            f.append(x)
        else:
            f.append("_")
    x = "".join(f)
    x = x.replace("__", "_").replace("__", "_")
    return x.strip("_").lower()


def __rename(model, name, value):
    name = name.split(":")[-1].strip()
    namespace = model.split("/")[0]
    return __python_name(namespace), __python_name(name), value


def __create_object(mapping):
    object = ModelCategory()
    for namespace, models in mapping.items():
        namespace_enum = ModelProvider()
        for name, value in models:
            setattr(namespace_enum, name, Model(**value))
        setattr(object, namespace, namespace_enum)
    return object


def __to_enum_object(models):
    f = [__rename(m["model"], m["name"], m) for m in models]
    mapping = {}
    for namespace, model, value in f:
        v = mapping.get(namespace, list())
        v.append((model, value))
        mapping[namespace] = v
    return __create_object(mapping)


def to_model_enum(models):
    if "chat" in models and "image" in models:  # v1
        chat_models = __to_enum_object(models["chat"])
        try:
            image_models = __to_enum_object(models["image"])
        except:
            image_models = __to_enum_object(models["image"][0])

        return chat_models, image_models
    else:  # v0
        return __to_enum_object(models)



import tempfile
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def make_temp_file(filename, content):
    with tempfile.TemporaryDirectory() as temp_directory:
        temp_file = Path(temp_directory) / filename
        with temp_file.open("w", encoding="utf-8") as writer:
            writer.write(content)
        yield temp_file



from json import loads


def youtube_trasncript_to_plain_text(text):
    objects = youtube_trasncript_to_objects(text)
    return " ".join((obj["text"] for obj in objects))


def youtube_trasncript_to_objects(text):
    return loads(text)



def _to_model_mapping_by_key(models, _key):
    if "chat" in models and "image" in models:  # v1
        mapping = {}
        for m in models["chat"]:
            key = m[_key]
            mapping[key] = m
        for m in models["image"]:
            key = m[_key]
            mapping[key] = m
        return mapping
    else:  # v0
        return dict([(m[_key], m) for m in models])


def to_model_mapping(models):
    return _to_model_mapping_by_key(models, "model")


def to_model_mapping_by_name(models):
    return _to_model_mapping_by_key(models, "name")


def __cheapest_model_v0(model_data):
    return sorted(model_data, key=lambda x: x["pricing"]["coins"])[0]


def __cheapest_model_v1(model_data):
    data = model_data["chat"]
    return sorted(data, key=lambda x: x["pricing"]["coins"])[0]


def cheapest_model(models):
    if "chat" in models and "image" in models:  # v1
        return __cheapest_model_v1(models)
    else:  # v0
        return __cheapest_model_v0(models)



from .api.v0_rag import SearchType


class AsyncStraicoRAG:
    def __init__(self, client, data):
        self.client = client
        self._data = data
        self.rag_id = data["_id"]

    @property
    def data(self):
        return self._data

    async def refresh(self):
        self._data = await self.client.rag(self.rag_id)

    async def delete(self) -> dict:
        return await self.client.rag_delete(self.rag_id)

    async def prompt_completion(
        self,
        model: str,
        message: str,
        search_type: [SearchType | str] = None,
        k: int = None,
        fetch_k: int = None,
        lambda_mult: float = None,
        score_threshold: float = None,
    ) -> dict:
        return await self.client.rag_prompt_completion(
            self.rag_id,
            model,
            message,
            search_type=search_type,
            k=k,
            fetch_k=fetch_k,
            lambda_mult=lambda_mult,
            score_threshold=score_threshold,
        )



import pytest
from .utils import make_temp_file
from aio_straico import straico_client, aio_straico_client


@pytest.mark.parametrize(
    "unsupported_file_name, content",
    [
        ("test.htm", """<html><body><p>Hello, World!</p></body></html"""),
        ("test.perl", """print "Hello, World!\n";"""),
    ],
)
def test_unsupported_file_upload(unsupported_file_name, content):
    with make_temp_file(unsupported_file_name, content) as file:
        with straico_client() as client:
            with pytest.raises(Exception, match="Unsupported File Type"):
                s = client.upload_file(file)


@pytest.mark.parametrize(
    "file_name, content",
    [
        # ('test.htm', """<html><body><p>Hello, World!</p></body></html"""),
        ("test.html", """<html><body><p>Hello, World!</p></body></html"""),
        ("test.json", """{"text": "Hello, World!"}"""),
        ("test.txt", """Hello, World!"""),
        ("test.csv", """Hello, World)"""),
        ("test.py", """print("Hello, World!")"""),
        ("test.php", """<?php echo "Hello, World!"; ?>"""),
        (
            "test.js",
            """function sayHello(): void { \nconsole.log("Hello, World!");\n}""",
        ),
        (
            "test.css",
            """body { \n    background-color: #f0f0f0; \n    font-family: Arial, sans-serif; \n}""",
        ),
        ("test.cs", """Console.WriteLine("Hello, World!");"""),
        ("test.swift", """import Foundation\n\nprint("Hello, World!")"""),
        ("test.kt", """fun main() {\n    println("Hello, World!")\n}"""),
        ("test.xml", """<test> hello world </test>"""),
        (
            "test.ts",
            """function sayHello(): void { \nconsole.log("Hello, World!");\n}""",
        ),
    ],
)
def test_supported_file_upload(file_name, content):
    with make_temp_file(file_name, content) as file:
        with straico_client() as client:
            s = client.upload_file(file)
            assert s is not None
            assert s.endswith("_" + file_name)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "unsupported_file_name, content",
    [
        ("test.htm", """<html><body><p>Hello, World!</p></body></html"""),
        ("test.perl", """print "Hello, World!\n";"""),
    ],
)
async def test_unsupported_file_upload_aio(unsupported_file_name, content):
    with make_temp_file(unsupported_file_name, content) as file:
        async with aio_straico_client() as client:
            with pytest.raises(Exception, match="Unsupported File Type"):
                s = await client.upload_file(file)
                print(s)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "file_name, content",
    [
        # ('test.htm', """<html><body><p>Hello, World!</p></body></html"""),
        ("test.html", """<html><body><p>Hello, World!</p></body></html"""),
        ("test.json", """{"text": "Hello, World!"}"""),
        ("test.txt", """Hello, World!"""),
        ("test.csv", """Hello, World)"""),
        ("test.py", """print("Hello, World!")"""),
        ("test.php", """<?php echo "Hello, World!"; ?>"""),
        (
            "test.js",
            """function sayHello(): void { \nconsole.log("Hello, World!");\n}""",
        ),
        (
            "test.css",
            """body { \n    background-color: #f0f0f0; \n    font-family: Arial, sans-serif; \n}""",
        ),
        ("test.cs", """Console.WriteLine("Hello, World!");"""),
        ("test.swift", """import Foundation\n\nprint("Hello, World!")"""),
        ("test.kt", """fun main() {\n    println("Hello, World!")\n}"""),
        ("test.xml", """<test> hello world </test>"""),
        (
            "test.ts",
            """function sayHello(): void { \nconsole.log("Hello, World!");\n}""",
        ),
    ],
)
async def test_supported_file_upload_aio(file_name, content):
    with make_temp_file(file_name, content) as file:
        async with aio_straico_client() as client:
            s = await client.upload_file(file)
            assert s is not None
            assert s.endswith("_" + file_name)



import pytest
from aio_straico import AsyncStraicoClient
from unittest.mock import patch, MagicMock


@pytest.mark.asyncio
async def test_async_client_initialization():
    client = AsyncStraicoClient()
    assert client is not None


@pytest.mark.asyncio
async def test_async_client_initialization_with_custom_params():
    client = AsyncStraicoClient(api_key="test_key", base_url="http://test.url")
    assert client.api_key == "test_key"
    assert client.base_url == "http://test.url"


@pytest.mark.asyncio
@patch("aio_straico.AsyncStraicoClient.some_async_method")
async def test_async_client_method_call(mock_method):
    client = AsyncStraicoClient()
    mock_method.return_value = "mocked response"
    response = await client.some_async_method()
    assert response == "mocked response"
    mock_method.assert_called_once()


@pytest.mark.asyncio
async def test_async_client_error_handling():
    client = AsyncStraicoClient()
    with pytest.raises(Exception) as e:
        await client.some_async_method()
    assert str(e.value) == "Expected error message"



import pathlib

import setuptools

# REFERENCE VIDEO : https://www.youtube.com/watch?v=WGsMydFFPMk

setuptools.setup(
    name="aio_straico",
    version="0.1.8",
    description="An unofficial async/sync client library for Straico API",
    long_description=pathlib.Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/jayrinaldime/aio-straico",
    author="Jay Rinaldi",
    author_email="jrinaldi@jayrinaldi.me",
    project_urls={
        "Source": "https://github.com/jayrinaldime/aio-straico",
        "Documentation": "https://github.com/jayrinaldime/aio-straico?tab=readme-ov-file#readme",
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: Freely Distributable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Communications :: Chat",
        "Topic :: Adaptive Technologies",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.9,<=3.13",
    install_requires=["httpx>=0.28.0"],
    packages=setuptools.find_packages(),
    include_package_data=True,
)
# Build step
# python -m build
# Push to pypi
# python -m twine upload --repository testpypi dist/*



from enum import Enum
from ..utils.tracing import observe, tracing_context, TRACING_ENABLED


async def aio_user(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/user"
    response = await session.get(url, headers=headers, **settings)
    return response


def user(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/user"
    response = session.get(url, headers=headers, **settings)
    return response


async def aio_models(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/models"
    response = await session.get(url, headers=headers, **settings)
    return response


def models(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/models"
    response = session.get(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
async def aio_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    model: str,
    message,
    temperature: float = None,
    max_tokens: float = None,
    **settings,
):
    url = f"{base_url}/v0/prompt/completion"
    json_body = {"model": model, "message": message}

    if "timeout" not in settings:
        settings["timeout"] = 60

    if temperature is not None:
        temperature = max(min(temperature, 2), 0)
        json_body["temperature"] = temperature

    if max_tokens is not None:
        max_tokens = max(max_tokens, 0)
        if max_tokens > 0:
            json_body["max_tokens"] = max_tokens
    if TRACING_ENABLED:
        tracing = dict(json_body)
        del tracing["model"]
        del tracing["message"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=model, model_parameters=tracing
        )
    response = await session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["data"]["completion"])
            del meta["choices"]
            tracing_context.update_current_observation(
                output=json_data["data"]["completion"]["choices"],
                usage_details={
                    "input": json_data["data"]["words"]["input"],
                    "output": json_data["data"]["words"]["output"],
                    "total": json_data["data"]["words"]["total"],
                    "input_cost": json_data["data"]["price"]["input"],
                    "output_cost": json_data["data"]["price"]["output"],
                    "total_cost": json_data["data"]["price"]["total"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )

    return response


@observe(as_type="generation")
def prompt_completion(
    session,
    base_url: str,
    headers: dict,
    model: str,
    message,
    temperature: float = None,
    max_tokens: float = None,
    **settings,
):
    url = f"{base_url}/v0/prompt/completion"
    json_body = {"model": model, "message": message}

    if "timeout" not in settings:
        settings["timeout"] = 60

    if temperature is not None:
        temperature = max(min(temperature, 2), 0)
        json_body["temperature"] = temperature

    if max_tokens is not None:
        max_tokens = max(max_tokens, 0)
        if max_tokens > 0:
            json_body["max_tokens"] = max_tokens
    if TRACING_ENABLED:
        tracing = dict(json_body)
        del tracing["model"]
        del tracing["message"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=model, model_parameters=tracing
        )
    response = session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["data"]["completion"])
            del meta["choices"]
            tracing_context.update_current_observation(
                output=json_data["data"]["completion"]["choices"],
                usage_details={
                    "input": json_data["data"]["words"]["input"],
                    "output": json_data["data"]["words"]["output"],
                    "total": json_data["data"]["words"]["total"],
                    "input_cost": json_data["data"]["price"]["input"],
                    "output_cost": json_data["data"]["price"]["output"],
                    "total_cost": json_data["data"]["price"]["total"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


async def aio_file_upload(
    session,
    base_url: str,
    headers: dict,
    *,
    filename: str,
    content_type: str,
    binary_data,
    **settings,
):
    files = {"file": (filename, binary_data, content_type)}

    url = f"{base_url}/v0/file/upload"
    if "timeout" not in settings:
        settings["timeout"] = 150

    response = await session.post(url, headers=headers, files=files, **settings)
    return response


def file_upload(
    session,
    base_url: str,
    headers: dict,
    *,
    filename: str,
    content_type: str,
    binary_data,
    **settings,
):
    files = {"file": (filename, binary_data, content_type)}

    url = f"{base_url}/v0/file/upload"
    if "timeout" not in settings:
        settings["timeout"] = 150
    response = session.post(url, headers=headers, files=files, **settings)
    return response


class ImageSize(Enum):
    square = "square"
    portrait = "portrait"
    landscape = "landscape"


def ImageSizer(size):
    lsize = size.lower()
    if lsize == ImageSize.square.value:
        return ImageSize.square
    elif lsize == ImageSize.portrait.value:
        return ImageSize.portrait
    elif lsize == ImageSize.landscape.value:
        return ImageSize.landscape

    raise Exception(f"Unknown Image Size {size}")


@observe(as_type="generation")
async def aio_image_generation(
    session,
    base_url: str,
    headers: dict,
    *,
    model: str,
    description: str,
    size: ImageSize | str,
    variations: int,
    seed: int = None,
    **settings,
):
    url = f"{base_url}/v0/image/generation"

    if not (0 < variations <= 4):
        raise Exception(f"Error variation size should be 1 to 4 got {variations}")

    size_type = type(size)
    if size_type == str and size not in [
        ImageSize.square.value,
        ImageSize.portrait.value,
        ImageSize.landscape.value,
    ]:
        raise Exception(f"Unknown Image Size {size}")
    elif size_type == ImageSize:
        size = size.value

    json_body = {
        "model": model,
        "description": description,
        "size": size,
        "variations": variations,
    }

    if seed is not None:
        json_body["seed"] = seed

    if "timeout" not in settings:
        settings["timeout"] = 300

    if TRACING_ENABLED:
        tracing = {"size": size, "variations": variations, **settings}
        tracing_context.update_current_observation(
            input=description, model=model, model_parameters=tracing
        )
    response = await session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            output = dict(json_data["data"])
            del output["price"]
            tracing_context.update_current_observation(
                output=output,
                usage_details={
                    "total_cost": json_data["data"]["price"]["total"],
                },
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


@observe(as_type="generation")
def image_generation(
    session,
    base_url: str,
    headers: dict,
    *,
    model: str,
    description: str,
    size: ImageSize | str,
    variations: int,
    seed: int = None,
    **settings,
):
    url = f"{base_url}/v0/image/generation"

    if not (0 < variations <= 4):
        raise Exception(f"Error variation size should be 1 to 4 got {variations}")

    size_type = type(size)
    if size_type == str and size not in [
        ImageSize.square.value,
        ImageSize.portrait.value,
        ImageSize.landscape.value,
    ]:
        raise Exception(f"Unknown Image Size {size}")
    elif size_type == ImageSize:
        size = size.value

    json_body = {
        "model": model,
        "description": description,
        "size": size,
        "variations": variations,
    }
    if seed is not None:
        json_body["seed"] = seed

    if "timeout" not in settings:
        settings["timeout"] = 300

    if TRACING_ENABLED:
        tracing = {"size": size, "variations": variations, **settings}
        tracing_context.update_current_observation(
            input=description, model=model, model_parameters=tracing
        )
    response = session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            output = dict(json_data["data"])
            del output["price"]
            tracing_context.update_current_observation(
                output=output,
                usage_details={
                    "total_cost": json_data["data"]["price"]["total"],
                },
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response



from enum import Enum
from pathlib import Path
from typing import List
from ..utils.tracing import observe, tracing_context, TRACING_ENABLED

valid_file_types = ("pdf", "docx", "csv", "txt", "xlsx", "py")


class ChunkingMethod(Enum):
    fixed_size = "fixed_size"
    # chunk_size = 1000
    # chunk_overlap = 50
    # separator = "\n" only single value is accepted
    recursive = "recursive"
    # chunk_size = 1000
    # chunk_overlap = 50
    # separator = ["\n\n", "\n", " ", ""] multiple values are accepted
    markdown = "markdown"
    # chunk_size = 1000
    # chunk_overlap = 50
    python = "python"
    # chunk_size = 1000
    # chunk_overlap = 50
    semantic = "semantic"
    # breakpoint_threshold_type: BreakpointThresholdType = percentile
    # buffer_size = 500


class BreakpointThresholdType(Enum):
    percentile = "percentile"
    interquartile = "interquartile"
    standard_deviation = "standard_deviation"
    gradient = "gradient"


class SearchType(Enum):
    similarity = "similarity"
    mmr = "mmr"
    similarity_score_threshold = "similarity_score_threshold"


async def aio_create_rag(
    session,
    base_url: str,
    headers: dict,
    name: str,
    description: str,
    files: [Path | str] = None,
    chunking_method: [ChunkingMethod | str] = None,
    chunk_size: int = 1000,
    chunk_overlap: int = 50,
    breakpoint_threshold_type: [
        BreakpointThresholdType | str
    ] = BreakpointThresholdType.percentile,
    buffer_size: int = 500,
    separator: [List[str] | str] = None,
    **settings,
):
    url = f"{base_url}/v0/rag"

    if "timeout" not in settings:
        settings["timeout"] = 600

    filepaths = [file if isinstance(file, Path) else Path(file) for file in files]
    files_parameter = []

    for index, file in enumerate(filepaths):
        extension = file.name.split(".")[-1].strip().lower()
        if extension not in valid_file_types:
            raise Exception(
                f"Unsupported file type {extension} for file {str(file)}. Only the following file types are supported {valid_file_types}"
            )

        if not file.exists():
            raise FileNotFoundError(str(file))

        if not file.is_file() or file.is_dir():
            raise Exception(f"Not a FILE {file}")

        with file.open("rb") as reader:
            content = reader.read(-1)
            files_parameter.append(
                ("files", (file.name, content, "application/octet-stream"))
            )

    payload = {"name": name, "description": description}
    if chunking_method is None:
        chunking_method = ChunkingMethod.fixed_size

    try:
        chunking_method = ChunkingMethod(chunking_method)
    except:
        raise Exception(f"Unexpected Chunking Method value {chunking_method}")

    if chunking_method == ChunkingMethod.fixed_size and separator is not None:
        if isinstance(separator, str):
            payload["separator"] = separator
        else:
            raise Exception(
                "Invalid value for separator. For Chunking method fixed size, only string is valid."
            )

    if chunking_method == ChunkingMethod.recursive and separator is not None:
        if isinstance(separator, list) or isinstance(separator, tuple):
            payload["separators"] = separator
        else:
            raise Exception(
                "Invalid value for separator. For Chunking method recursive, only list type is valid."
            )

    if chunking_method in (
        ChunkingMethod.fixed_size,
        ChunkingMethod.recursive,
        ChunkingMethod.markdown,
        ChunkingMethod.python,
    ):
        payload["chunk_size"] = chunk_size
        payload["chunk_overlap"] = chunk_overlap

    elif chunking_method == ChunkingMethod.semantic:
        payload["breakpoint_threshold_type"] = breakpoint_threshold_type
        payload["buffer_size"] = buffer_size

    payload["chunking_method"] = chunking_method.value

    response = await session.post(
        url, headers=headers, files=files_parameter, data=payload, **settings
    )
    return response


async def aio_rags(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/rag/user"
    response = await session.get(url, headers=headers, **settings)
    return response


async def aio_rag(session, base_url: str, headers: dict, rag_id: str, **settings):
    url = f"{base_url}/v0/rag/{rag_id}"
    response = await session.get(url, headers=headers, **settings)
    return response


async def aio_rag_delete(
    session, base_url: str, headers: dict, rag_id: str, **settings
):
    url = f"{base_url}/v0/rag/{rag_id}"
    response = await session.delete(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
async def aio_rag_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    rag_id: str,
    model: str,
    message: str,
    search_type: [SearchType | str] = None,
    k: int = None,
    fetch_k: int = None,
    lambda_mult: float = None,
    score_threshold: float = None,
    **settings,
):
    url = f"{base_url}/v0/rag/{rag_id}/prompt"
    payload = {
        "prompt": message,
        "model": model,
    }

    if "timeout" not in settings:
        settings["timeout"] = 600

    if search_type is not None:
        search_type = SearchType(search_type)
        payload["search_type"] = search_type.value

    if k is not None:
        payload["k"] = k

    if fetch_k is not None:
        payload["fetch_k"] = fetch_k

    if lambda_mult is not None:
        payload["lambda_mult"] = lambda_mult

    if score_threshold is not None:
        payload["score_threshold"] = score_threshold
    if TRACING_ENABLED:
        tracing = dict(payload)
        del tracing["model"]
        del tracing["prompt"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=model, model_parameters=tracing
        )
    response = await session.post(url, headers=headers, data=payload, **settings)
    if TRACING_ENABLED:
        if response.status_code == 200 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["response"])
            del meta["answer"]
            del meta["coins_used"]
            # del meta["overall_words"]
            # del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["response"]["answer"],
                usage_details={
                    "total_cost": json_data["response"]["coins_used"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


###########################
# Non Async Functions
#########################


def create_rag(
    session,
    base_url: str,
    headers: dict,
    name: str,
    description: str,
    files: [Path | str] = None,
    chunking_method: [ChunkingMethod | str] = None,
    chunk_size: int = 1000,
    chunk_overlap: int = 50,
    breakpoint_threshold_type: [
        BreakpointThresholdType | str
    ] = BreakpointThresholdType.percentile,
    buffer_size: int = 500,
    separator: [List[str] | str] = None,
    **settings,
):
    url = f"{base_url}/v0/rag"

    if "timeout" not in settings:
        settings["timeout"] = 600

    filepaths = [file if isinstance(file, Path) else Path(file) for file in files]
    if not all((file.exists() for file in filepaths)):
        pass

    files_parameter = []

    for index, file in enumerate(filepaths):
        extension = file.name.split(".")[-1].strip().lower()
        if extension not in valid_file_types:
            raise Exception(
                f"Unsupported file type {extension} for file {str(file)}. Only the following file types are supported {valid_file_types}"
            )

        if not file.exists():
            raise FileNotFoundError(str(file))

        if not file.is_file() or file.is_dir():
            raise Exception(f"Not a FILE {file}")

        with file.open("rb") as reader:
            content = reader.read(-1)
            files_parameter.append(
                ("files", (file.name, content, "application/octet-stream"))
            )

    payload = {"name": name, "description": description}

    if chunking_method is None:
        chunking_method = ChunkingMethod.fixed_size

    try:
        chunking_method = ChunkingMethod(chunking_method)
    except:
        raise Exception(f"Unexpected Chunking Method value {chunking_method}")

    if chunking_method == ChunkingMethod.fixed_size and separator is not None:
        if isinstance(separator, str):
            payload["separator"] = separator
        else:
            raise Exception(
                "Invalid value for separator. For Chunking method fixed size, only string is valid."
            )

    if chunking_method == ChunkingMethod.recursive and separator is not None:
        if isinstance(separator, list) or isinstance(separator, tuple):
            payload["separators"] = separator
        else:
            raise Exception(
                "Invalid value for separator. For Chunking method recursive, only list type is valid."
            )

    if chunking_method in (
        ChunkingMethod.fixed_size,
        ChunkingMethod.recursive,
        ChunkingMethod.markdown,
        ChunkingMethod.python,
    ):
        payload["chunk_size"] = chunk_size
        payload["chunk_overlap"] = chunk_overlap

    elif chunking_method == ChunkingMethod.semantic:
        payload["breakpoint_threshold_type"] = breakpoint_threshold_type
        payload["buffer_size"] = buffer_size

    payload["chunking_method"] = chunking_method.value

    response = session.post(
        url, headers=headers, files=files_parameter, data=payload, **settings
    )
    return response


def rags(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/rag/user"
    response = session.get(url, headers=headers, **settings)
    return response


def rag(session, base_url: str, headers: dict, rag_id: str, **settings):
    url = f"{base_url}/v0/rag/{rag_id}"
    response = session.get(url, headers=headers, **settings)
    return response


def rag_delete(session, base_url: str, headers: dict, rag_id: str, **settings):
    url = f"{base_url}/v0/rag/{rag_id}"
    response = session.delete(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
def rag_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    rag_id: str,
    model: str,
    message: str,
    search_type: [SearchType | str] = None,
    k: int = None,
    fetch_k: int = None,
    lambda_mult: float = None,
    score_threshold: float = None,
    **settings,
):
    url = f"{base_url}/v0/rag/{rag_id}/prompt"
    payload = {
        "prompt": message,
        "model": model,
    }

    if "timeout" not in settings:
        settings["timeout"] = 600

    if search_type is not None:
        search_type = SearchType(search_type)
        payload["search_type"] = search_type.value

    if k is not None:
        payload["k"] = k

    if fetch_k is not None:
        payload["fetch_k"] = fetch_k

    if lambda_mult is not None:
        payload["lambda_mult"] = lambda_mult

    if score_threshold is not None:
        payload["score_threshold"] = score_threshold
    if TRACING_ENABLED:
        tracing = dict(payload)
        del tracing["model"]
        del tracing["prompt"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=model, model_parameters=tracing
        )
    response = session.post(url, headers=headers, data=payload, **settings)
    if TRACING_ENABLED:
        if response.status_code == 200 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["response"])
            del meta["answer"]
            del meta["coins_used"]
            # del meta["overall_words"]
            # del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["response"]["answer"],
                usage_details={
                    "total_cost": json_data["response"]["coins_used"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response



from os import environ

TRACING_ENABLED = environ.get("STRAICO_TRACING_ENABLED", "False").lower() == "true"
if TRACING_ENABLED:
    from langfuse.decorators import observe
    from langfuse.decorators import langfuse_context as tracing_context

    def tracing_flush():
        tracing_context.flush()

else:
    from functools import wraps
    from inspect import iscoroutinefunction, isfunction

    def observe(*args, **kwargs):
        def decorator(func):
            return func

        if len(args) == 1 and (isfunction(args[0]) or iscoroutinefunction(args[0])):
            return args[0]
        return decorator

    class tracing_context:
        @classmethod
        def update_current_observation(cls, *args, **kwargs):
            return

    def tracing_flush():
        return None



from ..utils.tracing import observe, tracing_context, TRACING_ENABLED


async def aio_models(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v1/models"
    response = await session.get(url, headers=headers, **settings)
    return response


def models(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v1/models"
    response = session.get(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
async def aio_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    models,
    message,
    *,
    file_urls=[],
    images=[],
    youtube_urls=[],
    display_transcripts=False,
    temperature: float = None,
    max_tokens: float = None,
    **settings,
):
    if isinstance(models, str):
        models = [models]

    url = f"{base_url}/v1/prompt/completion"
    json_body = {"models": models, "message": message}

    if 0 < len(file_urls) <= 4:
        json_body["file_urls"] = file_urls

    if 0 < len(images) <= 4:
        json_body["images"] = images

    if 0 < len(youtube_urls) <= 4:
        json_body["youtube_urls"] = youtube_urls

    if display_transcripts:
        json_body["display_transcripts"] = True

    if "timeout" not in settings:
        settings["timeout"] = 300

    if temperature is not None:
        temperature = max(min(temperature, 2), 0)
        json_body["temperature"] = temperature

    if max_tokens is not None:
        max_tokens = max(max_tokens, 0)
        if max_tokens > 0:
            json_body["max_tokens"] = max_tokens
    if TRACING_ENABLED:
        tracing = dict(json_body)
        del tracing["models"]
        del tracing["message"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=", ".join(models), model_parameters=tracing
        )
    response = await session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["data"])
            del meta["completions"]
            del meta["overall_words"]
            del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["data"]["completions"],
                usage_details={
                    "input": json_data["data"]["overall_words"]["input"],
                    "output": json_data["data"]["overall_words"]["output"],
                    "total": json_data["data"]["overall_words"]["total"],
                    "input_cost": json_data["data"]["overall_price"]["input"],
                    "output_cost": json_data["data"]["overall_price"]["output"],
                    "total_cost": json_data["data"]["overall_price"]["total"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


@observe(as_type="generation")
def prompt_completion(
    session,
    base_url: str,
    headers: dict,
    models,
    message,
    *,
    file_urls=[],
    images=[],
    youtube_urls=[],
    display_transcripts=False,
    temperature: float = None,
    max_tokens: float = None,
    **settings,
):
    if isinstance(models, str):
        models = [models]

    url = f"{base_url}/v1/prompt/completion"
    json_body = {"models": models, "message": message}

    if 0 < len(file_urls) <= 4:
        json_body["file_urls"] = file_urls

    if 0 < len(images) <= 4:
        json_body["images"] = images

    if 0 < len(youtube_urls) <= 4:
        json_body["youtube_urls"] = youtube_urls

    if display_transcripts:
        json_body["display_transcripts"] = True

    if "timeout" not in settings:
        settings["timeout"] = 300

    if temperature is not None:
        temperature = max(min(temperature, 2), 0)
        json_body["temperature"] = temperature

    if max_tokens is not None:
        max_tokens = max(max_tokens, 0)
        if max_tokens > 0:
            json_body["max_tokens"] = max_tokens
    if TRACING_ENABLED:
        tracing = dict(json_body)
        del tracing["models"]
        del tracing["message"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=", ".join(models), model_parameters=tracing
        )
    response = session.post(url, headers=headers, json=json_body, **settings)
    if TRACING_ENABLED:
        if response.status_code == 201 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["data"])
            del meta["completions"]
            del meta["overall_words"]
            del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["data"]["completions"],
                usage_details={
                    "input": json_data["data"]["overall_words"]["input"],
                    "output": json_data["data"]["overall_words"]["output"],
                    "total": json_data["data"]["overall_words"]["total"],
                    "input_cost": json_data["data"]["overall_price"]["input"],
                    "output_cost": json_data["data"]["overall_price"]["output"],
                    "total_cost": json_data["data"]["overall_price"]["total"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response



from .api.v0_rag import SearchType
from .client_rag import StraicoRAG


class StraicoAgent:
    def __init__(self, client, data):
        self.client = client
        self._data = data
        self.agent_id = data["_id"]

    @property
    def data(self):
        return self._data

    def refresh(self):
        self._data = self.client.agent(self.agent_id)

    def delete(self) -> dict:
        return self.client.agent_delete(self.agent_id)

    def set_rag(self, rag: [StraicoRAG | dict | str]):
        return self.client.agent_add_rag(self.agent_id, rag)

    def prompt_completion(
        self,
        message: str,
        search_type: [SearchType | str] = None,
        k: int = None,
        fetch_k: int = None,
        lambda_mult: float = None,
        score_threshold: float = None,
    ) -> dict:
        return self.client.agent_prompt_completion(
            self.agent_id,
            message,
            search_type=search_type,
            k=k,
            fetch_k=fetch_k,
            lambda_mult=lambda_mult,
            score_threshold=score_threshold,
        )

    def update(
        self,
        *,
        rag: [StraicoRAG | dict | str] = None,
        name: str = None,
        description: str = None,
        model: str = None,
        system_prompt: str = None,
        tags: [str] = None,
    ):
        return self.client.agent_update(
            self.agent_id,
            name=name,
            rag=rag,
            description=description,
            model=model,
            system_prompt=system_prompt,
            tags=tags,
        )



from enum import Enum
from pathlib import Path

from .v0_rag import SearchType
from ..utils.tracing import observe, tracing_context, TRACING_ENABLED


async def aio_create_agent(
    session,
    base_url: str,
    headers: dict,
    name: str,
    description: str,
    model: str,
    system_prompt: str,
    tags: [str],
    **settings,
):
    url = f"{base_url}/v0/agent"

    payload = {
        "name": name,
        "description": description,
        "default_llm": model,
        "custom_prompt": system_prompt,
        "tags": tags,
    }

    response = await session.post(url, headers=headers, data=payload, **settings)
    return response


async def aio_add_rag_to_agent(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    rag_id: str,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}/rag"

    payload = {
        "rag": rag_id,
    }

    response = await session.post(url, headers=headers, data=payload, **settings)
    return response


async def aio_agents(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/agent"
    response = await session.get(url, headers=headers, **settings)
    return response


async def aio_agent(session, base_url: str, headers: dict, agent_id: str, **settings):
    url = f"{base_url}/v0/agent/{agent_id}"
    response = await session.get(url, headers=headers, **settings)
    return response


async def aio_agent_delete(
    session, base_url: str, headers: dict, agent_id: str, **settings
):
    url = f"{base_url}/v0/agent/{agent_id}"
    response = await session.delete(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
async def aio_agent_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    message: str,
    search_type: [SearchType | str] = None,
    k: int = None,
    fetch_k: int = None,
    lambda_mult: float = None,
    score_threshold: float = None,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}/prompt"
    payload = {
        "prompt": message,
    }

    if "timeout" not in settings:
        settings["timeout"] = 600

    if search_type is not None:
        search_type = SearchType(search_type)
        payload["search_type"] = search_type.value

    if k is not None:
        payload["k"] = k

    if fetch_k is not None:
        payload["fetch_k"] = fetch_k

    if lambda_mult is not None:
        payload["lambda_mult"] = lambda_mult

    if score_threshold is not None:
        payload["score_threshold"] = score_threshold
    if TRACING_ENABLED:
        tracing = dict(payload)
        del tracing["prompt"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=agent_id, model_parameters=tracing
        )
    response = await session.post(url, headers=headers, data=payload, **settings)
    if TRACING_ENABLED:
        if response.status_code == 200 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["response"])
            del meta["answer"]
            del meta["coins_used"]
            # del meta["overall_words"]
            # del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["response"]["answer"],
                usage_details={
                    "total_cost": json_data["response"]["coins_used"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


async def aio_agent_update(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    *,
    rag_id: str = None,
    name: str = None,
    description: str = None,
    model: str = None,
    system_prompt: str = None,
    tags: [str] = None,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}"

    payload = {}
    if rag_id is not None:
        payload["rag"] = rag_id
    if name is not None:
        payload["name"] = name
    if description is not None:
        payload["description"] = description
    if system_prompt is not None:
        payload["custom_prompt"] = system_prompt
    if model is not None:
        payload["default_llm"] = model
    if tags is not None:
        payload["tags"] = tags

    response = await session.put(url, headers=headers, data=payload, **settings)
    return response


def create_agent(
    session,
    base_url: str,
    headers: dict,
    name: str,
    description: str,
    model: str,
    system_prompt: str,
    tags: [str],
    **settings,
):
    url = f"{base_url}/v0/agent"

    payload = {
        "name": name,
        "description": description,
        "default_llm": model,
        "custom_prompt": system_prompt,
        "tags": tags,
    }

    response = session.post(url, headers=headers, data=payload, **settings)
    return response


def add_rag_to_agent(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    rag_id: str,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}/rag"

    payload = {
        "rag": rag_id,
    }

    response = session.post(url, headers=headers, data=payload, **settings)
    return response


def agents(session, base_url: str, headers: dict, **settings):
    url = f"{base_url}/v0/agent"
    response = session.get(url, headers=headers, **settings)
    return response


def agent(session, base_url: str, headers: dict, agent_id: str, **settings):
    url = f"{base_url}/v0/agent/{agent_id}"
    response = session.get(url, headers=headers, **settings)
    return response


def agent_delete(session, base_url: str, headers: dict, agent_id: str, **settings):
    url = f"{base_url}/v0/agent/{agent_id}"
    response = session.delete(url, headers=headers, **settings)
    return response


@observe(as_type="generation")
def agent_prompt_completion(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    message: str,
    search_type: [SearchType | str] = None,
    k: int = None,
    fetch_k: int = None,
    lambda_mult: float = None,
    score_threshold: float = None,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}/prompt"
    payload = {
        "prompt": message,
    }

    if "timeout" not in settings:
        settings["timeout"] = 600

    if search_type is not None:
        search_type = SearchType(search_type)
        payload["search_type"] = search_type.value

    if k is not None:
        payload["k"] = k

    if fetch_k is not None:
        payload["fetch_k"] = fetch_k

    if lambda_mult is not None:
        payload["lambda_mult"] = lambda_mult

    if score_threshold is not None:
        payload["score_threshold"] = score_threshold
    if TRACING_ENABLED:
        tracing = dict(payload)
        del tracing["prompt"]
        tracing.update(settings)
        tracing_context.update_current_observation(
            input=message, model=agent_id, model_parameters=tracing
        )
    response = session.post(url, headers=headers, data=payload, **settings)
    if TRACING_ENABLED:
        if response.status_code == 200 and response.json()["success"]:
            json_data = response.json()
            meta = dict(json_data["response"])
            del meta["answer"]
            del meta["coins_used"]
            # del meta["overall_words"]
            # del meta["overall_price"]
            tracing_context.update_current_observation(
                output=json_data["response"]["answer"],
                usage_details={
                    "total_cost": json_data["response"]["coins_used"],
                },
                metadata=meta,
                status_message=str(response.status_code),
            )

        else:
            tracing_context.update_current_observation(
                output=response.text, status_message=str(response.status_code)
            )
    return response


def agent_update(
    session,
    base_url: str,
    headers: dict,
    agent_id: str,
    *,
    rag_id: str = None,
    name: str = None,
    description: str = None,
    model: str = None,
    system_prompt: str = None,
    tags: [str] = None,
    **settings,
):
    url = f"{base_url}/v0/agent/{agent_id}"

    payload = {}
    if rag_id is not None:
        payload["rag"] = rag_id
    if name is not None:
        payload["name"] = name
    if description is not None:
        payload["description"] = description
    if system_prompt is not None:
        payload["custom_prompt"] = system_prompt
    if model is not None:
        payload["default_llm"] = model
    if tags is not None:
        payload["tags"] = tags

    response = session.put(url, headers=headers, data=payload, **settings)
    return response



from .models import cheapest_model, to_model_mapping, to_model_mapping_by_name

from .transcript_utils import (
    youtube_trasncript_to_objects,
    youtube_trasncript_to_plain_text,
)

from .models_to_enum import to_model_enum


from collections.abc import Iterable


def is_listable_not_string(obj):
    return not isinstance(obj, str) and isinstance(obj, Iterable)
